/* istanbul ignore file */
/* tslint:disable */
/* eslint-disable */

export type Distributions = {
    distribution: string;
    params: Record<string, (number | string | Array<number> | Array<string>)>;
};
