/* istanbul ignore file */
/* tslint:disable */
/* eslint-disable */

export type TensorboardConfig = {
    id?: number;
    sweep_id: string;
    shared_folder: string;
    stage?: string;
    desired_stage?: string;
    url?: string;
};
